@@ -16,7 +16,6 @@
 # ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 # OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
-import os
 import unittest
 import binascii
 
@@ -27,10 +26,12 @@ import dns.name
 import dns.rdataclass
 import dns.rdatatype
 import dns.rrset
+import dns.tsig
 import dns.update
+import dns.rdtypes.ANY.OPT
+import dns.rdtypes.ANY.TSIG
 
-def here(filename):
-    return os.path.join(os.path.dirname(__file__), filename)
+from tests.util import here
 
 query_text = """id 1234
 opcode QUERY
@@ -338,6 +339,11 @@ class MessageTestCase(unittest.TestCase):
         self.assertEqual(m.edns, 0)
         self.assertTrue(m.ednsflags & dns.flags.DO)
 
+    def test_EDNS_default_payload_is_1232(self):
+        m = dns.message.make_query('foo', 'A')
+        m.use_edns()
+        self.assertEqual(m.payload, dns.message.DEFAULT_EDNS_PAYLOAD)
+
     def test_from_file(self):
         m = dns.message.from_file(here('query'))
         expected = dns.message.from_text(query_text)
@@ -365,6 +371,13 @@ class MessageTestCase(unittest.TestCase):
         q.additional = [rrset]
         self.assertEqual(q.sections[3], [rrset])
 
+    def test_is_a_response_empty_question(self):
+        q = dns.message.make_query('www.dnspython.org.', 'a')
+        r = dns.message.make_response(q)
+        r.question = []
+        r.set_rcode(dns.rcode.FORMERR)
+        self.assertTrue(q.is_response(r))
+
     def test_not_a_response(self):
         q = dns.message.QueryMessage(id=1)
         self.assertFalse(q.is_response(q))
@@ -381,6 +394,11 @@ class MessageTestCase(unittest.TestCase):
         q2.id = 1
         r = dns.message.make_response(q2)
         self.assertFalse(q1.is_response(r))
+        # Now set rcode to FORMERR and check again.  It should still
+        # not be a response as we check the question section for FORMERR
+        # if it is present.
+        r.set_rcode(dns.rcode.FORMERR)
+        self.assertFalse(q1.is_response(r))
         # Test the other case of differing questions, where there is
         # something in the response's question section that is not in
         # the question's.  We have to do multiple questions to test
@@ -419,6 +437,16 @@ class MessageTestCase(unittest.TestCase):
         self.assertEqual(q.payload, 4096)
         self.assertEqual(q.options, ())
 
+    def test_setting_id(self):
+        q = dns.message.make_query('www.dnspython.org.', 'a', id=12345)
+        self.assertEqual(q.id, 12345)
+
+    def test_setting_flags(self):
+        q = dns.message.make_query('www.dnspython.org.', 'a',
+                                   flags=dns.flags.RD|dns.flags.CD)
+        self.assertEqual(q.flags, dns.flags.RD|dns.flags.CD)
+        self.assertEqual(q.flags, 0x0110)
+
     def test_generic_message_class(self):
         q1 = dns.message.Message(id=1)
         q1.set_opcode(dns.opcode.NOTIFY)
@@ -433,5 +461,274 @@ class MessageTestCase(unittest.TestCase):
         self.assertFalse(isinstance(q2, dns.update.UpdateMessage))
         self.assertEqual(q1, q2)
 
+    def test_truncated_exception_message(self):
+        q = dns.message.Message(id=1)
+        q.flags |= dns.flags.TC
+        te = dns.message.Truncated(message=q)
+        self.assertEqual(te.message(), q)
+
+    def test_bad_opt(self):
+        # Not in addtional
+        q = dns.message.Message(id=1)
+        opt = dns.rdtypes.ANY.OPT.OPT(1200, dns.rdatatype.OPT, ())
+        rrs = dns.rrset.from_rdata(dns.name.root, 0, opt)
+        q.answer.append(rrs)
+        wire = q.to_wire()
+        with self.assertRaises(dns.message.BadEDNS):
+            dns.message.from_wire(wire)
+        # Owner name not root name
+        q = dns.message.Message(id=1)
+        rrs = dns.rrset.from_rdata('foo.', 0, opt)
+        q.additional.append(rrs)
+        wire = q.to_wire()
+        with self.assertRaises(dns.message.BadEDNS):
+            dns.message.from_wire(wire)
+        # Multiple opts
+        q = dns.message.Message(id=1)
+        rrs = dns.rrset.from_rdata(dns.name.root, 0, opt)
+        q.additional.append(rrs)
+        q.additional.append(rrs)
+        wire = q.to_wire()
+        with self.assertRaises(dns.message.BadEDNS):
+            dns.message.from_wire(wire)
+
+    def test_bad_tsig(self):
+        keyname = dns.name.from_text('key.')
+        # Not in addtional
+        q = dns.message.Message(id=1)
+        tsig = dns.rdtypes.ANY.TSIG.TSIG(dns.rdataclass.ANY, dns.rdatatype.TSIG,
+                                         dns.tsig.HMAC_SHA256, 0, 300, b'1234',
+                                         0, 0, b'')
+        rrs = dns.rrset.from_rdata(keyname, 0, tsig)
+        q.answer.append(rrs)
+        wire = q.to_wire()
+        with self.assertRaises(dns.message.BadTSIG):
+            dns.message.from_wire(wire)
+        # Multiple tsigs
+        q = dns.message.Message(id=1)
+        q.additional.append(rrs)
+        q.additional.append(rrs)
+        wire = q.to_wire()
+        with self.assertRaises(dns.message.BadTSIG):
+            dns.message.from_wire(wire)
+        # Class not ANY
+        tsig = dns.rdtypes.ANY.TSIG.TSIG(dns.rdataclass.IN, dns.rdatatype.TSIG,
+                                         dns.tsig.HMAC_SHA256, 0, 300, b'1234',
+                                         0, 0, b'')
+        rrs = dns.rrset.from_rdata(keyname, 0, tsig)
+        wire = q.to_wire()
+        with self.assertRaises(dns.message.BadTSIG):
+            dns.message.from_wire(wire)
+
+    def test_read_no_content_message(self):
+        m = dns.message.from_text(';comment')
+        self.assertIsInstance(m, dns.message.QueryMessage)
+
+    def test_eflags_turns_on_edns(self):
+        m = dns.message.from_text('eflags DO')
+        self.assertIsInstance(m, dns.message.QueryMessage)
+        self.assertEqual(m.edns, 0)
+
+    def test_payload_turns_on_edns(self):
+        m = dns.message.from_text('payload 1200')
+        self.assertIsInstance(m, dns.message.QueryMessage)
+        self.assertEqual(m.payload, 1200)
+
+    def test_bogus_header(self):
+        with self.assertRaises(dns.message.UnknownHeaderField):
+            dns.message.from_text('bogus foo')
+
+    def test_question_only(self):
+        m = dns.message.from_text(answer_text)
+        w = m.to_wire()
+        r = dns.message.from_wire(w, question_only=True)
+        self.assertEqual(r.id, m.id)
+        self.assertEqual(r.question[0], m.question[0])
+        self.assertEqual(len(r.answer), 0)
+        self.assertEqual(len(r.authority), 0)
+        self.assertEqual(len(r.additional), 0)
+
+    def test_bad_resolve_chaining(self):
+        r = dns.message.make_query('www.dnspython.org.', 'a')
+        with self.assertRaises(dns.message.NotQueryResponse):
+            r.resolve_chaining()
+        r.flags |= dns.flags.QR
+        r.id = 1
+        r.find_rrset(r.question, dns.name.from_text('example'),
+                     dns.rdataclass.IN, dns.rdatatype.A, create=True,
+                     force_unique=True)
+        with self.assertRaises(dns.exception.FormError):
+            r.resolve_chaining()
+
+    def test_resolve_chaining_no_infinite_loop(self):
+        r = dns.message.from_text('''id 1
+flags QR
+;QUESTION
+www.example. IN CNAME
+;AUTHORITY
+example. 300 IN SOA . . 1 2 3 4 5
+''')
+        # passing is not going into an infinite loop in this call
+        result = r.resolve_chaining()
+        self.assertEqual(result.canonical_name,
+                         dns.name.from_text('www.example.'))
+        self.assertEqual(result.minimum_ttl, 5)
+        self.assertIsNone(result.answer)
+
+    def test_bad_text_questions(self):
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+;QUESTION
+example.
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+;QUESTION
+example. IN
+''')
+        with self.assertRaises(dns.rdatatype.UnknownRdatatype):
+            dns.message.from_text('''id 1
+;QUESTION
+example. INA
+''')
+        with self.assertRaises(dns.rdatatype.UnknownRdatatype):
+            dns.message.from_text('''id 1
+;QUESTION
+example. IN BOGUS
+''')
+
+    def test_bad_text_rrs(self):
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+example. IN A
+;ANSWER
+example.
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+example. IN A
+;ANSWER
+example. IN
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+example. IN A
+;ANSWER
+example. 300
+''')
+        with self.assertRaises(dns.rdatatype.UnknownRdatatype):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+example. IN A
+;ANSWER
+example. 30a IN A
+''')
+        with self.assertRaises(dns.rdatatype.UnknownRdatatype):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+example. IN A
+;ANSWER
+example. 300 INA A
+''')
+        with self.assertRaises(dns.exception.UnexpectedEnd):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+example. IN A
+;ANSWER
+example. 300 IN A
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+opcode UPDATE
+;ZONE
+example. IN SOA
+;UPDATE
+example. 300 IN A
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+opcode UPDATE
+;ZONE
+example. IN SOA
+;UPDATE
+example. 300 NONE A
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+opcode UPDATE
+;ZONE
+example. IN SOA
+;PREREQ
+example. 300 NONE A 10.0.0.1
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+;ANSWER
+            300 IN A 10.0.0.1
+''')
+        with self.assertRaises(dns.exception.SyntaxError):
+            dns.message.from_text('''id 1
+flags QR
+;QUESTION
+            IN SOA
+''')
+
+    def test_from_wire_makes_Flag(self):
+        m = dns.message.from_wire(goodwire)
+        self.assertIsInstance(m.flags, dns.flags.Flag)
+        self.assertEqual(m.flags, dns.flags.Flag.RD)
+
+    def test_continue_on_error(self):
+        good_message = dns.message.from_text(
+"""id 1234
+opcode QUERY
+rcode NOERROR
+flags QR AA RD
+;QUESTION
+www.dnspython.org. IN SOA
+;ANSWER
+www.dnspython.org. 300 IN SOA . . 1 2 3 4 4294967295
+www.dnspython.org. 300 IN A 1.2.3.4
+www.dnspython.org. 300 IN AAAA ::1
+""")
+        wire = good_message.to_wire()
+        # change ANCOUNT to 255
+        bad_wire = wire[:6] + b'\x00\xff' + wire[8:]
+        # change AAAA into rdata with rdlen 0
+        bad_wire = bad_wire[:-18] + b'\x00' * 2
+        m = dns.message.from_wire(bad_wire, continue_on_error=True)
+        self.assertEqual(len(m.errors), 2)
+        print(m.errors)
+        self.assertEqual(str(m.errors[0].exception),
+                         'IPv6 addresses are 16 bytes long')
+        self.assertEqual(str(m.errors[1].exception),
+                         'DNS message is malformed.')
+        expected_message = dns.message.from_text(
+"""id 1234
+opcode QUERY
+rcode NOERROR
+flags QR AA RD
+;QUESTION
+www.dnspython.org. IN SOA
+;ANSWER
+www.dnspython.org. 300 IN SOA . . 1 2 3 4 4294967295
+www.dnspython.org. 300 IN A 1.2.3.4
+""")
+        self.assertEqual(m, expected_message)
+
+
 if __name__ == '__main__':
     unittest.main()
