@@ -32,11 +32,11 @@ import dns.rdataset
 import dns.rdataclass
 import dns.rdatatype
 import dns.rrset
+import dns.versioned
 import dns.zone
 import dns.node
 
-def here(filename):
-    return os.path.join(os.path.dirname(__file__), filename)
+from tests.util import here
 
 example_text = """$TTL 3600
 $ORIGIN example.
@@ -145,7 +145,7 @@ Königsgäßchen 300 NS Königsgäßchen
 misc_cases_input = """
 $ORIGIN example.
 $TTL 300
-      
+
 @ soa foo bar 1 2 3 4 5
 @ ns ns1
 @ ns ns2
@@ -175,6 +175,72 @@ $ORIGIN example.
 @ 300 ns ns2
 """
 
+example_comments_text = """$TTL 3600
+$ORIGIN example.
+@ soa foo bar (1 ; not kept
+2 3 4 5) ; kept
+@ ns ns1
+@ ns ns2
+ns1 a 10.0.0.1 ; comment1
+ns2 a 10.0.0.2 ; comment2
+"""
+
+example_comments_text_output = """@ 3600 IN SOA foo bar 1 2 3 4 5 ; kept
+@ 3600 IN NS ns1
+@ 3600 IN NS ns2
+ns1 3600 IN A 10.0.0.1 ; comment1
+ns2 3600 IN A 10.0.0.2 ; comment2
+"""
+
+
+example_cname = """$TTL 3600
+$ORIGIN example.
+@ soa foo bar (1 2 3 4 5)
+@ ns ns1
+@ ns ns2
+ns1 a 10.0.0.1
+ns2 a 10.0.0.2
+www a 10.0.0.3
+web cname www
+    nsec @ CNAME RRSIG
+    rrsig NSEC 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+    rrsig CNAME 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+web2 cname www
+    nsec3 1 1 12 aabbccdd 2t7b4g4vsa5smi47k61mv5bv1a22bojr CNAME RRSIG
+    rrsig NSEC3 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+    rrsig CNAME 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+"""
+
+
+example_other_data = """$TTL 3600
+$ORIGIN example.
+@ soa foo bar (1 2 3 4 5)
+@ ns ns1
+@ ns ns2
+ns1 a 10.0.0.1
+ns2 a 10.0.0.2
+www a 10.0.0.3
+web a 10.0.0.4
+    nsec @ A RRSIG
+    rrsig A 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+    rrsig NSEC 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+"""
+
+example_cname_and_other_data = """$TTL 3600
+$ORIGIN example.
+@ soa foo bar (1 2 3 4 5)
+@ ns ns1
+@ ns ns2
+ns1 a 10.0.0.1
+ns2 a 10.0.0.2
+www a 10.0.0.3
+web a 10.0.0.4
+    cname www
+    nsec @ A RRSIG
+    rrsig A 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+    rrsig NSEC 1 3 3600 20200101000000 20030101000000 2143 foo MxFcby9k/yvedMfQgKzhH5er0Mu/vILz 45IkskceFGgiWCn/GxHhai6VAuHAoNUz 4YoU1tVfSCSqQYn6//11U6Nld80jEeC8 aTrO+KKmCaY=
+"""
+
 _keep_output = True
 
 def _rdata_sort(a):
@@ -217,7 +283,7 @@ def compare_files(test_name, a_name, b_name):
 
 class ZoneTestCase(unittest.TestCase):
 
-    def testFromFile1(self): # type: () -> None
+    def testFromFile1(self):
         z = dns.zone.from_file(here('example'), 'example')
         ok = False
         try:
@@ -230,7 +296,7 @@ class ZoneTestCase(unittest.TestCase):
                 os.unlink(here('example1.out'))
         self.assertTrue(ok)
 
-    def testFromFile2(self): # type: () -> None
+    def testFromFile2(self):
         z = dns.zone.from_file(here('example'), 'example', relativize=False)
         ok = False
         try:
@@ -243,7 +309,7 @@ class ZoneTestCase(unittest.TestCase):
                 os.unlink(here('example2.out'))
         self.assertTrue(ok)
 
-    def testToFileTextualStream(self): # type: () -> None
+    def testToFileTextualStream(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         f = StringIO()
         z.to_file(f)
@@ -251,7 +317,7 @@ class ZoneTestCase(unittest.TestCase):
         f.close()
         self.assertEqual(out, example_text_output)
 
-    def testToFileBinaryStream(self): # type: () -> None
+    def testToFileBinaryStream(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         f = BytesIO()
         z.to_file(f, nl=b'\n')
@@ -259,7 +325,7 @@ class ZoneTestCase(unittest.TestCase):
         f.close()
         self.assertEqual(out, example_text_output.encode())
 
-    def testToFileTextual(self): # type: () -> None
+    def testToFileTextual(self):
         z = dns.zone.from_file(here('example'), 'example')
         try:
             f = open(here('example3-textual.out'), 'w')
@@ -273,7 +339,7 @@ class ZoneTestCase(unittest.TestCase):
                 os.unlink(here('example3-textual.out'))
         self.assertTrue(ok)
 
-    def testToFileBinary(self): # type: () -> None
+    def testToFileBinary(self):
         z = dns.zone.from_file(here('example'), 'example')
         try:
             f = open(here('example3-binary.out'), 'wb')
@@ -287,7 +353,7 @@ class ZoneTestCase(unittest.TestCase):
                 os.unlink(here('example3-binary.out'))
         self.assertTrue(ok)
 
-    def testToFileFilename(self): # type: () -> None
+    def testToFileFilename(self):
         z = dns.zone.from_file(here('example'), 'example')
         try:
             z.to_file(here('example3-filename.out'))
@@ -299,7 +365,7 @@ class ZoneTestCase(unittest.TestCase):
                 os.unlink(here('example3-filename.out'))
         self.assertTrue(ok)
 
-    def testToText(self): # type: () -> None
+    def testToText(self):
         z = dns.zone.from_file(here('example'), 'example')
         ok = False
         try:
@@ -315,7 +381,35 @@ class ZoneTestCase(unittest.TestCase):
                 os.unlink(here('example3.out'))
         self.assertTrue(ok)
 
-    def testFromText(self): # type: () -> None
+    def testToFileTextualWithOrigin(self):
+        z = dns.zone.from_file(here('example'), 'example')
+        try:
+            f = open(here('example4-textual.out'), 'w')
+            z.to_file(f, want_origin=True)
+            f.close()
+            ok = compare_files('testToFileTextualWithOrigin',
+                               here('example4-textual.out'),
+                               here('example4.good'))
+        finally:
+            if not _keep_output:
+                os.unlink(here('example4-textual.out'))
+        self.assertTrue(ok)
+
+    def testToFileBinaryWithOrigin(self):
+        z = dns.zone.from_file(here('example'), 'example')
+        try:
+            f = open(here('example4-binary.out'), 'wb')
+            z.to_file(f, want_origin=True)
+            f.close()
+            ok = compare_files('testToFileBinaryWithOrigin',
+                               here('example4-binary.out'),
+                               here('example4.good'))
+        finally:
+            if not _keep_output:
+                os.unlink(here('example4-binary.out'))
+        self.assertTrue(ok)
+
+    def testFromText(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         f = StringIO()
         names = list(z.nodes.keys())
@@ -325,7 +419,7 @@ class ZoneTestCase(unittest.TestCase):
             f.write('\n')
         self.assertEqual(f.getvalue(), example_text_output)
 
-    def testTorture1(self): # type: () -> None
+    def testTorture1(self):
         #
         # Read a zone containing all our supported RR types, and
         # for each RR in the zone, convert the rdata into wire format
@@ -346,151 +440,167 @@ class ZoneTestCase(unittest.TestCase):
                                               origin=o)
                     self.assertEqual(rd, rd2)
 
-    def testEqual(self): # type: () -> None
+    def testEqual(self):
         z1 = dns.zone.from_text(example_text, 'example.', relativize=True)
         z2 = dns.zone.from_text(example_text_output, 'example.',
                                 relativize=True)
         self.assertEqual(z1, z2)
 
-    def testNotEqual1(self): # type: () -> None
+    def testNotEqual1(self):
         z1 = dns.zone.from_text(example_text, 'example.', relativize=True)
         z2 = dns.zone.from_text(something_quite_similar, 'example.',
                                 relativize=True)
         self.assertNotEqual(z1, z2)
 
-    def testNotEqual2(self): # type: () -> None
+    def testNotEqual2(self):
         z1 = dns.zone.from_text(example_text, 'example.', relativize=True)
         z2 = dns.zone.from_text(something_different, 'example.',
                                 relativize=True)
         self.assertNotEqual(z1, z2)
 
-    def testNotEqual3(self): # type: () -> None
+    def testNotEqual3(self):
         z1 = dns.zone.from_text(example_text, 'example.', relativize=True)
         z2 = dns.zone.from_text(something_different, 'example2.',
                                 relativize=True)
         self.assertNotEqual(z1, z2)
 
-    def testFindRdataset1(self): # type: () -> None
+    def testFindRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rds = z.find_rdataset('@', 'soa')
         exrds = dns.rdataset.from_text('IN', 'SOA', 300, 'foo bar 1 2 3 4 5')
         self.assertEqual(rds, exrds)
 
-    def testFindRdataset2(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testFindRdataset2(self):
+        def bad():
             z = dns.zone.from_text(example_text, 'example.', relativize=True)
             z.find_rdataset('@', 'loc')
         self.assertRaises(KeyError, bad)
 
-    def testFindRRset1(self): # type: () -> None
+    def testFindRRset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rrs = z.find_rrset('@', 'soa')
         exrrs = dns.rrset.from_text('@', 300, 'IN', 'SOA', 'foo bar 1 2 3 4 5')
         self.assertEqual(rrs, exrrs)
 
-    def testFindRRset2(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testFindRRset2(self):
+        def bad():
             z = dns.zone.from_text(example_text, 'example.', relativize=True)
             z.find_rrset('@', 'loc')
         self.assertRaises(KeyError, bad)
 
-    def testGetRdataset1(self): # type: () -> None
+    def testGetRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rds = z.get_rdataset('@', 'soa')
         exrds = dns.rdataset.from_text('IN', 'SOA', 300, 'foo bar 1 2 3 4 5')
         self.assertEqual(rds, exrds)
 
-    def testGetRdataset2(self): # type: () -> None
+    def testGetRdataset2(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rds = z.get_rdataset('@', 'loc')
         self.assertTrue(rds is None)
 
-    def testGetRRset1(self): # type: () -> None
+    def testGetRRset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rrs = z.get_rrset('@', 'soa')
         exrrs = dns.rrset.from_text('@', 300, 'IN', 'SOA', 'foo bar 1 2 3 4 5')
         self.assertEqual(rrs, exrrs)
 
-    def testGetRRset2(self): # type: () -> None
+    def testGetRRset2(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rrs = z.get_rrset('@', 'loc')
         self.assertTrue(rrs is None)
 
-    def testReplaceRdataset1(self): # type: () -> None
+    def testReplaceRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rdataset = dns.rdataset.from_text('in', 'ns', 300, 'ns3', 'ns4')
         z.replace_rdataset('@', rdataset)
         rds = z.get_rdataset('@', 'ns')
         self.assertTrue(rds is rdataset)
 
-    def testReplaceRdataset2(self): # type: () -> None
+    def testReplaceRdataset2(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         rdataset = dns.rdataset.from_text('in', 'txt', 300, '"foo"')
         z.replace_rdataset('@', rdataset)
         rds = z.get_rdataset('@', 'txt')
         self.assertTrue(rds is rdataset)
 
-    def testDeleteRdataset1(self): # type: () -> None
+    def testDeleteRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         z.delete_rdataset('@', 'ns')
         rds = z.get_rdataset('@', 'ns')
         self.assertTrue(rds is None)
 
-    def testDeleteRdataset2(self): # type: () -> None
+    def testDeleteRdataset2(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         z.delete_rdataset('ns1', 'a')
         node = z.get_node('ns1')
         self.assertTrue(node is None)
 
-    def testNodeFindRdataset1(self): # type: () -> None
+    def testNodeFindRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         node = z['@']
         rds = node.find_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA)
         exrds = dns.rdataset.from_text('IN', 'SOA', 300, 'foo bar 1 2 3 4 5')
         self.assertEqual(rds, exrds)
 
-    def testNodeFindRdataset2(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testNodeFindRdataset2(self):
+        def bad():
             z = dns.zone.from_text(example_text, 'example.', relativize=True)
             node = z['@']
             node.find_rdataset(dns.rdataclass.IN, dns.rdatatype.LOC)
         self.assertRaises(KeyError, bad)
 
-    def testNodeGetRdataset1(self): # type: () -> None
+    def testNodeFindRdataset3(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True)
+        node = z['@']
+        rds = node.find_rdataset(dns.rdataclass.IN, dns.rdatatype.RRSIG,
+                                 dns.rdatatype.A, create=True)
+        self.assertEqual(rds.rdclass, dns.rdataclass.IN)
+        self.assertEqual(rds.rdtype, dns.rdatatype.RRSIG)
+        self.assertEqual(rds.covers, dns.rdatatype.A)
+
+    def testNodeGetRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         node = z['@']
         rds = node.get_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA)
         exrds = dns.rdataset.from_text('IN', 'SOA', 300, 'foo bar 1 2 3 4 5')
         self.assertEqual(rds, exrds)
 
-    def testNodeGetRdataset2(self): # type: () -> None
+    def testNodeGetRdataset2(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         node = z['@']
         rds = node.get_rdataset(dns.rdataclass.IN, dns.rdatatype.LOC)
         self.assertTrue(rds is None)
 
-    def testNodeDeleteRdataset1(self): # type: () -> None
+    def testNodeDeleteRdataset1(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         node = z['@']
         node.delete_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA)
         rds = node.get_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA)
         self.assertTrue(rds is None)
 
-    def testNodeDeleteRdataset2(self): # type: () -> None
+    def testNodeDeleteRdataset2(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         node = z['@']
         node.delete_rdataset(dns.rdataclass.IN, dns.rdatatype.LOC)
         rds = node.get_rdataset(dns.rdataclass.IN, dns.rdatatype.LOC)
         self.assertTrue(rds is None)
 
-    def testIterateRdatasets(self): # type: () -> None
+    def testIterateNodes(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True)
+        count = 0
+        for n in z:
+            count += 1
+        self.assertEqual(count, 4)
+
+    def testIterateRdatasets(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         ns = [n for n, r in z.iterate_rdatasets('A')]
         ns.sort()
         self.assertEqual(ns, [dns.name.from_text('ns1', None),
                               dns.name.from_text('ns2', None)])
 
-    def testIterateAllRdatasets(self): # type: () -> None
+    def testIterateAllRdatasets(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         ns = [n for n, r in z.iterate_rdatasets()]
         ns.sort()
@@ -500,7 +610,7 @@ class ZoneTestCase(unittest.TestCase):
                               dns.name.from_text('ns1', None),
                               dns.name.from_text('ns2', None)])
 
-    def testIterateRdatas(self): # type: () -> None
+    def testIterateRdatas(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         l = list(z.iterate_rdatas('A'))
         l.sort()
@@ -514,7 +624,7 @@ class ZoneTestCase(unittest.TestCase):
                                     '10.0.0.2'))]
         self.assertEqual(l, exl)
 
-    def testIterateAllRdatas(self): # type: () -> None
+    def testIterateAllRdatas(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
         l = list(z.iterate_rdatas())
         l.sort(key=_rdata_sort)
@@ -556,17 +666,17 @@ class ZoneTestCase(unittest.TestCase):
         self.assertEqual(z.get('foo'), n)
         del z['foo']
         self.assertEqual(z.get('foo'), None)
-        def bad1():
+        with self.assertRaises(KeyError):
             z[123] = n
-        self.assertRaises(KeyError, bad1)
-        def bad2():
+        with self.assertRaises(KeyError):
             z['foo.'] = n
-        self.assertRaises(KeyError, bad2)
-        def bad3():
+        with self.assertRaises(KeyError):
             bn = z.find_node('bar')
-        self.assertRaises(KeyError, bad3)
         bn = z.find_node('bar', True)
         self.assertTrue(isinstance(bn, dns.node.Node))
+        # The next two tests pass by not raising KeyError
+        z.delete_node('foo')
+        z.delete_node('bar')
 
     def testBadReplacement(self):
         z = dns.zone.from_text(example_text, 'example.', relativize=True)
@@ -575,7 +685,7 @@ class ZoneTestCase(unittest.TestCase):
             z.replace_rdataset('foo', rds)
         self.assertRaises(ValueError, bad)
 
-    def testTTLs(self): # type: () -> None
+    def testTTLs(self):
         z = dns.zone.from_text(ttl_example_text, 'example.', relativize=True)
         n = z['@'] # type: dns.node.Node
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA))
@@ -587,7 +697,7 @@ class ZoneTestCase(unittest.TestCase):
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.A))
         self.assertEqual(rds.ttl, 694861)
 
-    def testTTLFromSOA(self): # type: () -> None
+    def testTTLFromSOA(self):
         z = dns.zone.from_text(ttl_from_soa_text, 'example.', relativize=True)
         n = z['@']
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA))
@@ -600,7 +710,7 @@ class ZoneTestCase(unittest.TestCase):
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.A))
         self.assertEqual(rds.ttl, soa_rd.minimum)
 
-    def testTTLFromLast(self): # type: () -> None
+    def testTTLFromLast(self):
         z = dns.zone.from_text(ttl_from_last_text, 'example.', check_origin=False)
         n = z['@']
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.NS))
@@ -612,33 +722,33 @@ class ZoneTestCase(unittest.TestCase):
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.A))
         self.assertEqual(rds.ttl, 694861)
 
-    def testNoTTL(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testNoTTL(self):
+        def bad():
             dns.zone.from_text(no_ttl_text, 'example.', check_origin=False)
         self.assertRaises(dns.exception.SyntaxError, bad)
 
-    def testNoSOA(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testNoSOA(self):
+        def bad():
             dns.zone.from_text(no_soa_text, 'example.', relativize=True)
         self.assertRaises(dns.zone.NoSOA, bad)
 
-    def testNoNS(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testNoNS(self):
+        def bad():
             dns.zone.from_text(no_ns_text, 'example.', relativize=True)
         self.assertRaises(dns.zone.NoNS, bad)
 
-    def testInclude(self): # type: () -> None
+    def testInclude(self):
         z1 = dns.zone.from_text(include_text, 'example.', relativize=True,
                                 allow_include=True)
         z2 = dns.zone.from_file(here('example'), 'example.', relativize=True)
         self.assertEqual(z1, z2)
 
-    def testBadDirective(self): # type: () -> None
-        def bad(): # type: () -> None
+    def testBadDirective(self):
+        def bad():
             dns.zone.from_text(bad_directive_text, 'example.', relativize=True)
         self.assertRaises(dns.exception.SyntaxError, bad)
 
-    def testFirstRRStartsWithWhitespace(self): # type: () -> None
+    def testFirstRRStartsWithWhitespace(self):
         # no name is specified, so default to the initial origin
         z = dns.zone.from_text(' 300 IN A 10.0.0.1', origin='example.',
                                check_origin=False)
@@ -646,21 +756,21 @@ class ZoneTestCase(unittest.TestCase):
         rds = cast(dns.rdataset.Rdataset, n.get_rdataset(dns.rdataclass.IN, dns.rdatatype.A))
         self.assertEqual(rds.ttl, 300)
 
-    def testZoneOrigin(self): # type: () -> None
+    def testZoneOrigin(self):
         z = dns.zone.Zone('example.')
         self.assertEqual(z.origin, dns.name.from_text('example.'))
-        def bad1(): # type: () -> None
+        def bad1():
             o = dns.name.from_text('example', None)
             dns.zone.Zone(o)
         self.assertRaises(ValueError, bad1)
-        def bad2(): # type: () -> None
+        def bad2():
             dns.zone.Zone(cast(str, 1.0))
         self.assertRaises(ValueError, bad2)
 
-    def testZoneOriginNone(self): # type: () -> None
+    def testZoneOriginNone(self):
         dns.zone.Zone(cast(str, None))
 
-    def testZoneFromXFR(self): # type: () -> None
+    def testZoneFromXFR(self):
         z1_abs = dns.zone.from_text(example_text, 'example.', relativize=False)
         z2_abs = dns.zone.from_xfr(make_xfr(z1_abs), relativize=False)
         self.assertEqual(z1_abs, z2_abs)
@@ -746,5 +856,205 @@ class ZoneTestCase(unittest.TestCase):
         self.assertEqual(z._validate_name('foo.bar.example.'),
                          dns.name.from_text('foo.bar', None))
 
+    def testComments(self):
+        z = dns.zone.from_text(example_comments_text, 'example.',
+                               relativize=True)
+        f = StringIO()
+        z.to_file(f, want_comments=True)
+        out = f.getvalue()
+        f.close()
+        self.assertEqual(out, example_comments_text_output)
+
+    def testUncomparable(self):
+        z = dns.zone.from_text(example_comments_text, 'example.',
+                               relativize=True)
+        self.assertFalse(z == 'a')
+
+    def testUnsorted(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True)
+        f = StringIO()
+        z.to_file(f, sorted=False)
+        out = f.getvalue()
+        f.close()
+        z2 = dns.zone.from_text(out, 'example.', relativize=True)
+        self.assertEqual(z, z2)
+
+    def testNodeReplaceRdatasetConvertsRRsets(self):
+        node = dns.node.Node()
+        rrs = dns.rrset.from_text('foo', 300, 'in', 'a', '10.0.0.1')
+        node.replace_rdataset(rrs)
+        rds = node.find_rdataset(dns.rdataclass.IN, dns.rdatatype.A)
+        self.assertEqual(rds, rrs)
+        self.assertTrue(rds is not rrs)
+        self.assertFalse(isinstance(rds, dns.rrset.RRset))
+
+    def testCnameAndOtherDataAddOther(self):
+        z = dns.zone.from_text(example_cname, 'example.', relativize=True)
+        rds = dns.rdataset.from_text('in', 'a', 300, '10.0.0.1')
+        z.replace_rdataset('web', rds)
+        z.replace_rdataset('web2', rds.copy())
+        n = z.find_node('web')
+        self.assertEqual(len(n.rdatasets), 3)
+        self.assertEqual(n.find_rdataset(dns.rdataclass.IN, dns.rdatatype.A),
+                         rds)
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.NSEC))
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.RRSIG,
+                                            dns.rdatatype.NSEC))
+        n = z.find_node('web2')
+        self.assertEqual(len(n.rdatasets), 3)
+        self.assertEqual(n.find_rdataset(dns.rdataclass.IN, dns.rdatatype.A),
+                         rds)
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.NSEC3))
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.RRSIG,
+                                            dns.rdatatype.NSEC3))
+
+    def testCnameAndOtherDataAddCname(self):
+        z = dns.zone.from_text(example_other_data, 'example.', relativize=True)
+        rds = dns.rdataset.from_text('in', 'cname', 300, 'www')
+        z.replace_rdataset('web', rds)
+        n = z.find_node('web')
+        self.assertEqual(len(n.rdatasets), 3)
+        self.assertEqual(n.find_rdataset(dns.rdataclass.IN,
+                                         dns.rdatatype.CNAME),
+                         rds)
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.NSEC))
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.RRSIG,
+                                            dns.rdatatype.NSEC))
+
+    def testCnameAndOtherDataInZonefile(self):
+        with self.assertRaises(dns.zonefile.CNAMEAndOtherData):
+            dns.zone.from_text(example_cname_and_other_data, 'example.',
+                               relativize=True)
+
+    def testNameInZoneWithStr(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=False)
+        self.assertTrue('ns1.example.' in z)
+        self.assertTrue('bar.foo.example.' in z)
+
+    def testNameInZoneWhereNameIsNotValid(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=False)
+        with self.assertRaises(KeyError):
+            self.assertTrue(1 in z)
+
+
+class VersionedZoneTestCase(unittest.TestCase):
+    def testUseTransaction(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        with self.assertRaises(dns.versioned.UseTransaction):
+            z.find_node('not_there', True)
+        with self.assertRaises(dns.versioned.UseTransaction):
+            z.delete_node('not_there')
+        with self.assertRaises(dns.versioned.UseTransaction):
+            z.find_rdataset('not_there', 'a', create=True)
+        with self.assertRaises(dns.versioned.UseTransaction):
+            z.get_rdataset('not_there', 'a', create=True)
+        with self.assertRaises(dns.versioned.UseTransaction):
+            z.delete_rdataset('not_there', 'a')
+        with self.assertRaises(dns.versioned.UseTransaction):
+            z.replace_rdataset('not_there', None)
+
+    def testImmutableNodes(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        node = z.find_node('@')
+        with self.assertRaises(TypeError):
+            node.find_rdataset(dns.rdataclass.IN, dns.rdatatype.RP,
+                               create=True)
+        with self.assertRaises(TypeError):
+            node.get_rdataset(dns.rdataclass.IN, dns.rdatatype.RP,
+                               create=True)
+        with self.assertRaises(TypeError):
+            node.delete_rdataset(dns.rdataclass.IN, dns.rdatatype.SOA)
+        with self.assertRaises(TypeError):
+            node.replace_rdataset(None)
+
+    def testSelectDefaultPruningPolicy(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        z.set_pruning_policy(None)
+        self.assertEqual(z._pruning_policy, z._default_pruning_policy)
+
+    def testSetAlternatePruningPolicyInConstructor(self):
+        def never_prune(version):
+            return False
+        z = dns.versioned.Zone('example', pruning_policy=never_prune)
+        self.assertEqual(z._pruning_policy, never_prune)
+
+    def testCannotSpecifyBothSerialAndVersionIdToReader(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        with self.assertRaises(ValueError):
+            z.reader(1, 1)
+
+    def testUnknownVersion(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        with self.assertRaises(KeyError):
+            z.reader(99999)
+
+    def testUnknownSerial(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        with self.assertRaises(KeyError):
+            z.reader(serial=99999)
+
+    def testNoRelativizeReader(self):
+        z = dns.zone.from_text(example_text, 'example.', relativize=False,
+                               zone_factory=dns.versioned.Zone)
+        with z.reader(serial=1) as txn:
+            rds = txn.get('example.', 'soa')
+            self.assertEqual(rds[0].serial, 1)
+
+    def testCnameAndOtherDataAddOther(self):
+        z = dns.zone.from_text(example_cname, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        rds = dns.rdataset.from_text('in', 'a', 300, '10.0.0.1')
+        with z.writer() as txn:
+            txn.replace('web', rds)
+            txn.replace('web2', rds.copy())
+        n = z.find_node('web')
+        self.assertEqual(len(n.rdatasets), 3)
+        self.assertEqual(n.find_rdataset(dns.rdataclass.IN, dns.rdatatype.A),
+                         rds)
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.NSEC))
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.RRSIG,
+                                            dns.rdatatype.NSEC))
+        n = z.find_node('web2')
+        self.assertEqual(len(n.rdatasets), 3)
+        self.assertEqual(n.find_rdataset(dns.rdataclass.IN, dns.rdatatype.A),
+                         rds)
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.NSEC3))
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.RRSIG,
+                                            dns.rdatatype.NSEC3))
+
+    def testCnameAndOtherDataAddCname(self):
+        z = dns.zone.from_text(example_other_data, 'example.', relativize=True,
+                               zone_factory=dns.versioned.Zone)
+        rds = dns.rdataset.from_text('in', 'cname', 300, 'www')
+        with z.writer() as txn:
+            txn.replace('web', rds)
+        n = z.find_node('web')
+        self.assertEqual(len(n.rdatasets), 3)
+        self.assertEqual(n.find_rdataset(dns.rdataclass.IN,
+                                         dns.rdatatype.CNAME),
+                         rds)
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.NSEC))
+        self.assertIsNotNone(n.get_rdataset(dns.rdataclass.IN,
+                                            dns.rdatatype.RRSIG,
+                                            dns.rdatatype.NSEC))
+
+
 if __name__ == '__main__':
     unittest.main()
